--[[
Copyright 2023 darkfrei

The MIT License
https://opensource.org/license/mit/

Permission is hereby granted, free of charge, to any person obtaining 
a copy of this software and associated documentation files (the “Software”), 
to deal in the Software without restriction, including without limitation 
the rights to use, copy, modify, merge, publish, distribute, sublicense, 
and/or sell copies of the Software, and to permit persons to whom 
the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included 
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, 
EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES 
OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, 
DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, 
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH 
THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
--]]


--vfa = require ('vfa')

local vfa = {}

local function sortXYBackward (list)
	table.sort(list, function(a, b) return a.y > b.y or a.y == b.y and a.x > b.x or false end)
end

local function sortX (list)
	table.sort(list, function(a, b) return a.x < b.x end)
end

local function reload ()
	vfa.dirY = 0
	vfa.parabolaLines = {}
	vfa.sweepLine = {}
	vfa.queue = {}
	for i = 1, #vfa.points-1, 2 do
		local x, y = vfa.points[i], vfa.points[i+1]
		table.insert (vfa.queue, {x=x, y=y, point=true})
	end
	sortXYBackward (vfa.queue)
	for i = #vfa.queue, 1, -1 do
		local event = vfa.queue[i]
--		print (i, event.x, event.y)
	end
end

local function getCircumcircle (x1, y1, x2, y2, x3, y3)
	local d = 2*(x1*(y2-y3)+x2*(y3-y1)+x3*(y1-y2))
--	print (d)
	local t1, t2, t3 = x1*x1+y1*y1, x2*x2+y2*y2, x3*x3+y3*y3
	local x = (t1*(y2-y3)+t2*(y3-y1)+t3*(y1-y2)) / d
	local y = (t1*(x3-x2)+t2*(x1-x3)+t3*(x2-x1)) / d
	local radius = math.sqrt((x1-x)^2+(y1-y)^2)
	return x, y, radius
end

function vfa.load ()
	vfa.points = {}
	vfa.points = {100,100, 320, 120, 400, 200}
	for i = 1, 10 do
		local x = math.random (Width-40)+20
		local y = math.random (Height*1-40)+20
		table.insert (vfa.points, x)
		table.insert (vfa.points, y)
--		table.insert (vfa.points, x+200)
--		table.insert (vfa.points, y)
	end
	reload ()
end

local function pointEvent (point)
	table.insert (vfa.sweepLine, point)
	sortX (vfa.sweepLine)
	local index
	for i, event in ipairs (vfa.sweepLine) do
		if event == point then
			index = i
			break
		end
	end

	if #vfa.sweepLine > 2 then
		local circleEvent = {}
--		print ('index', index)
		local p1 = vfa.sweepLine[index-2]
		local p2 = vfa.sweepLine[index-1]
		local p3 = vfa.sweepLine[index]
		local p4 = vfa.sweepLine[index+1]
		local p5 = vfa.sweepLine[index+2]
		if p1 and p2 and p3 then
			local x, y, r = getCircumcircle (p1.x, p1.y, p2.x, p2.y, p3.x, p3.y)
			local circle = {x=x, cy=y, y=y+r, r=r}
--			print ('circle', 'x:'..x,'y:'..y,'r:'..r)
			table.insert (vfa.queue, circle)
		end
		if p3 and p4 and p5 then
			local x, y, r = getCircumcircle (p3.x, p3.y, p4.x, p4.y, p5.x, p5.y)
			local circle = {x=x, cy=y, y=y+r, r=r}
--			print ('circle', 'x:'..x,'y:'..y,'r:'..r)
			table.insert (vfa.queue, circle)
		end
	end
end



local function circleEvent (point)
	
end

local function getParabolaCrossPoint (p1x, p1y, p2x, p2y, dirY) -- focus 1, focus 2, directrix
-- Function to find the intersection point of two parabolas
	local f1 = math.abs(dirY-p1y)/2
	local f2 = math.abs(dirY-p2y)/2

	local a1 = -1/(4*f1)
	local a2 = -1/(4*f2)
	local b1 = -2*p1x*a1
	local b2 = -2*p2x*a2
	local c1 = p1x*p1x*a1 + p1y + f1
	local c2 = p2x*p2x*a2 + p2y + f2
	local a = a1-a2
	local b = b1-b2
	local c = c1-c2

	local d = b*b-4*a*c
	local x, y
	if d >=0 then
		x = (-b-math.sqrt (d))/(2*a)
		y = a1*x*x + b1*x + c1
	end
	return x, y
end

local function getParabolaLine (x, y, dirY, xMin, xMax) -- focus, directrix, horizontal limits
-- Function to generate points along a parabola curve
	local line = {}
	local f = math.abs(dirY-y)/2
	local a = -1/(4*f) 
	local b = -2*x*a
	local c = x*x*a + y + f
	local nSteps = math.floor((xMax-xMin)/5)
	local step = (xMax-xMin)/nSteps
	for i = 0, nSteps do
		local x0 = xMin + i * step
		local y0 = a*x0*x0 + b*x0 + c
		table.insert (line, x0)
		table.insert (line, y0)
	end
	return line
end

function vfa.update (dt)
	vfa.dirY = vfa.dirY+0.25*60*dt
	if vfa.dirY > Height then
		reload ()
	end

	for i = #vfa.queue, 1, -1 do
		local event = vfa.queue[i]
		if vfa.dirY > event.y then
			table.remove (vfa.queue, i)
			if event.point then
				pointEvent (event)
			else
				circleEvent (event)
			end
			sortXYBackward (vfa.queue)
		else
			break
		end
	end
	
	-- update parabolas
	local leftX = 0
	vfa.parabolaLines = {}
	for i, focus in ipairs (vfa.sweepLine) do
		local focus2 = vfa.sweepLine[i+1]
		local rightX = Width
		local righty
		if not (i == #vfa.sweepLine) then
			rightX = getParabolaCrossPoint (focus.x, focus.y, focus2.x, focus2.y, vfa.dirY)
		end
		local line = getParabolaLine (focus.x, focus.y, vfa.dirY, leftX, rightX)
		if #line > 3 then
			table.insert (vfa.parabolaLines, line)
		end
		leftX = rightX
	end
end

function vfa.draw ()


	love.graphics.points (vfa.points)
	love.graphics.line (0, vfa.dirY, Width, vfa.dirY)

	
love.graphics.setColor (1,1,0)
	for i, event in ipairs (vfa.sweepLine) do
		love.graphics.circle ('line', event.x, event.y, 5)
	end
	
	love.graphics.setColor (1,0,0)
	
	for i, event in ipairs (vfa.queue) do
		if event.point then
			love.graphics.circle ('line', event.x, event.y, 3)
		else
			love.graphics.circle ('line', event.x, event.cy, event.r)
		end
	end
	
	love.graphics.setColor (1,1,1)
	for i, line in ipairs (vfa.parabolaLines) do
		love.graphics.line (line)
	end
end

-------------------------------------
-- love
-------------------------------------

love.window.setMode(400, 400)
Width, Height = love.graphics.getDimensions( )


function love.load()
	vfa.load ()
	pause = false
end


function love.update(dt)
	if not pause then
		vfa.update (dt)
	end
end


function love.draw()
	vfa.draw ()
end

function love.keypressed(key, scancode, isrepeat)
	if false then
elseif key == "space" then
	pause = not pause
	elseif key == "escape" then
		love.event.quit()
	end
end

function love.mousepressed( x, y, button, istouch, presses )
	if button == 1 then -- left mouse button
	elseif button == 2 then -- right mouse button
	end
end

function love.mousemoved( x, y, dx, dy, istouch )
end

function love.mousereleased( x, y, button, istouch, presses )
	if button == 1 then -- left mouse button
	elseif button == 2 then -- right mouse button
	end
end